import io
import json
import pandas as pd


def result_to_dataframe(json_dict):
    """ Deserializes a dataframe that was serialized with orient='split'
    """
    if 'column_names' and 'index_names' in json_dict.keys():
        columns = pd.MultiIndex.from_tuples(json_dict['columns'], names=json_dict['column_names'])
        index = pd.MultiIndex.from_tuples(json_dict['index'], names=json_dict['index_names'])
    else:
        columns = json_dict['columns']
        index = json_dict['index']
    df = pd.DataFrame(data=json_dict['data'])
    df.columns = columns
    df.index = index
    return df

def find_columns_in_crosstab(key, crosstab):
    columns = []
    if key in crosstab:
        if key == 'f':
            columns += list(crosstab[key].keys())
        elif isinstance(crosstab[key], str):
            columns.append(crosstab[key])
        elif isinstance(crosstab[key], list):
            columns = crosstab[key]
    return columns

def get_columns_from_crosstabs(crosstabs):
    columns = []
    for crosstab in crosstabs:
        for key in ['x', 'y', 'f', 'w']:
            columns += find_columns_in_crosstab(key, crosstab)
    return list(set(columns))



from tally.tally import Tally
from tally.dataset import DataSet
from tally.build import Build
from tally.sheet import Sheet
from tally.options import Options
from tally.helpers import result_to_dataframe



import json
import requests
import pandas as pd
import copy
from .decorators import verify_token

class Tally:
    """Tally is a wrapper for the cloud-based Tally API.

    Parameters
    ----------
    api_key : string
        API key used to authenticate calls. This is provided by Datasmoothie, email info@datasmoothie.com to get one.

    Attributes
    ----------
    __api_key : string
        The API key used for authentication.
    __headers : type
        The headers used for http requests.

    """

    def __init__(self, api_key, host="tally.datasmoothie.com/tally/", ssl=True):
        """Initialise the client with an API key.

        Parameters
        ----------
        api_key : string
            The API key used for authentication, provided by Datasmoothie.
        host : string
            The path to the server api
        ssl : boolean
            Datasmoothie does not support non ssl communication, but this is
            useful for development and local unit testing.

        """
        self.host = host
        if ssl == True or ssl == 'True':
            self.base_url = "https://{}".format(host)
        else:
            self.base_url = "http://{}".format(host)
        self.__api_key = api_key
        self.__headers = {
            "Authorization": "Token {}".format(self.__api_key),
            "Content-Type": "application/json",
            "Accept": "application/json"
            }

    def _get_headers(self):
        return self.__headers

    def get_request(self, resource, action=None):
        """Send a get request to the API with a convenient wrapper.

        Parameters
        ----------
        resource : string
            Name of the resource we are calling.
            These are the root paths of the API.
        action : type
            Name of the action to take on the resouce,
            e.g. datasource/1/meta_data
        binary : boolean
            Are we sending data in a binary format or not? Spss files are binary, CSVs are not.

        Returns
        -------
        type
            JSON object representing the result of the API request.

        """
        if action is None:
            action = ""
        request_path = "{}/{}/{}".format(self.base_url, resource, action)
        result = requests.get(request_path, headers=self._get_headers())
        result = json.loads(result.content)
        return result

    @verify_token
    def post_request(self, resource, action="", data={}, files=None):
        """Send a POST request to the API with a wrapper.

        This is used by other objects
        to call the API and not used by the users themselves.

        Parameters
        ----------
        resource : string
            Name of the resource we are calling.
            These are the root paths of the API.
        action : string
            Name of the action to take on the resouce,
            e.g. datasource/1/meta_data
        data : type
            JSON object with the payload to send with a POST request.

        Returns
        -------
        type
            Description of returned object.

        """
        # http doesn't support nested data structures, so we flatten the params
        if 'params' in data.keys():
            data['params'] = json.dumps(data['params'])
        if len(action) == 0:
            request_path = "{}/{}/".format(self.base_url, resource)
        else:
            request_path = "{}/{}/{}/".format(self.base_url, resource, action)
        if files is not None:
            headers = copy.deepcopy(self._get_headers())
            if "Content-Type" in headers.keys():
                del headers["Content-Type"]
            result = requests.post(request_path, headers=headers, data=data, files=files)
        else:
            result = requests.post(request_path, headers=self._get_headers(), data=json.dumps(data))
        return result

    def put_request(self, resource, data):
        request_path = "{}/{}/".format(self.base_url, resource)
        result = requests.put(request_path,
                              headers=self._get_headers(),
                              json=data
                              )
        return result

    def delete_request(self, resource, primary_key):
        """Send a delete request to the API.

        Parameters
        ----------
        resource : string
            Path location of the resource.
        primary_key : string
            The reports primary key, as reported by get reports.

        Returns
        -------
        type
            The response from the server.

        """
        request_path = "{}/{}/{}".format(self.base_url, resource, primary_key)
        result = requests.delete(request_path,
                                 headers=self._get_headers())
        return result

    def get_base_url(self, api=True):
        if api:
            return self.host
        else:
            return self.host.replace("api2", "")

    def build_excel(self, filename=None, params={}):
        """
        Build Excel tables with aggregated results.

        Parameters
        ----------
        filename : string
            Name of file that has the survey data. to be used 

        """


    def weight(self):
        """Get a list of all the datasources this account has.

        Returns
        -------
        type
            A JSON object with meta information about the datasources.
            This can be used to get the primary keys of the datasource
            the user wants to manipulate.

        """
        result = self.get_request('datasource')
        return result




import copy

from .decorators import verify_no_tables

class Options:
    """Options for both sheet and build with convenience methods to change both data and visuals. 
    """

    def __init__(self, table_options=None, formats=None, parent=None):
        if table_options is None:
            self.table_options = {
                "format":{
                    "base":{},
                    "percentage":{},
                    "counts":{}
                },
                "stub": {"ci":["c%"]},
                "banner_border": True,
                "annotations": False
            }
        else:
            self.table_options = table_options

        self.global_options = {
            "font_name": "Calibri",
            "font_size": "11"
        }

        self.formats = formats
        if formats is None:
            self.formats = {}
        self.parent = parent

    def set_base_position(self, position):
        """Change the position of the base

        Parameters
        ==========
        position: string
            Can be `outside`, `hidden` og `default`. Outside means it appears above the table's stub label.
        """
        if position == 'outside':
            self.set_question_format('base', {'font_color':'ffffff', 'bg_color':'ffffff', 'font_size':1})
        self.table_options['base'] = position

    def set_base_labels(self, label, unweighted_label="Unweighted base"):
        """Set the labels for the base and unweighted base.

        If this is not set, the bases will be called "weighted base" and "unweighted base". This can be used to change the base to "United Kingdom" or something that reflects the base.

        Parameters
        ==========
        label: string
            Main label for the base.
        unweighted_label: string
            Label for the unweighted base, if both are being shown.
        """

        self.table_options['base_label'] = label
        self.table_options['unweighted_base_label'] = unweighted_label

    def set_font(self, name, size=11):
        """Sets the global font name and size to be used in Excel.

        Parameters
        ==========
        name: string
            Font name
        size: int
            Font size
        """
        self.table_options['font_name'] = name
        self.global_options['font_name'] = name
        self.table_options['font_size'] = size
        self.global_options['font_size'] = size
        self.set_question_format('percentage', {'font_name':name, 'font_size':size})

    def set_language_key(self, text_key):
        """Set language key for results.

        Parameters
        ==========
        text_key: string
          Text key for the language, e.g. en-GB or jp-JP. The meta data has to have the available key present.
        """
        self.table_options['stub']['text_key'] = text_key

    def set_show_table_base_column(self, xtotal):
        """ Show column on the far left with the totals from all of the data.

        Parameters
        ==========
        xtotal: boolean
          Show/hide the total.
        """
        self.table_options['stub']['xtotal'] = xtotal

    def set_annotations(self, annotations):
        """Set whether tables are annotated with information about sig-testing, weight applied, and base.
       
        Parameters
        ==========
        annotations: boolean
          True if we want to see annotations under every table.
        """
        self.table_options['annotations'] = annotations

    def set_sig_test_levels(self,sig_test_level):
        """ Set alpha level for sig-testing.

        Parameters
        ==========
        sig_test_leve: int or array
            Integer or array of integers indicating the confidence level we want. No sig-tests are run if this isn't set.
        """
        if type(sig_test_level) == list:
            self.table_options['stub']['sig_level'] = sig_test_level
        else:
            self.table_options['stub']['sig_level'] = [sig_test_level]

    def set_hide_gridlines(self, hide):
        """Show or hide gridlines in the spreadsheet.

        Parameters
        ==========
        hide: integer
            0 Don't hide gridlines, 1 Hide printed gridlines only, 2 Hide screen and printed gridlines.
        """
        self._set_page_setup('hide_gridlines', {"option":hide})

    @verify_no_tables
    def set_stats(self, stats=['mean']):
        """Decide on what statistics to show
        """
        self.table_options['stub']['stats'] = stats

    @verify_no_tables
    def set_filter(self, filter):
        """Set filter to use on all data

        Parameters
        ==========
            filter: logical statement
                Logical statement, such as {'country':[1]} to use only country with code 1.
        """
        self.table_options['stub']['f'] = filter

    @verify_no_tables
    def set_ci(self, ci):
        """Set what the table cells should contain

        Parameters
        ==========
        ci: one or both of `c%`, `counts` to show percentages and/or counts.
        """
        self.table_options['stub']['ci'] = ci

    @verify_no_tables
    def set_show_bases(self, bases):
        """Set what bases should be shown in the table.

        Parameters
        ==========
        bases: list
            A list of strings, options are 'auto', 'weighted', 'unweighted', 'both', where auto will show the weighted base if there is a base, otherwise the unweighted one.
        """
        if bases not in ['auto', 'weighted', 'unweighted', 'both']:
            raise ValueError('Bases to show must be both, auto, weighted or unweighted')
        self.table_options['stub']['base'] = bases

    @verify_no_tables
    def set_stub(self, default_stub):
        """Set what variables should be in the stub.

        These can include any options used by DataSet.crosstab.
        """
        self.table_options['stub'] = {**self.table_options['stub'], **default_stub}

    @verify_no_tables
    def set_weight(self, weight):
        """Set the variable used to weight. 
        
        Every table will use this variable to weight its results if set.
        """
        self.table_options['stub']['w'] = weight

    @verify_no_tables
    def set_banner_border(self, border):
        """Set whether there should be a border between the banner categories, like gender and age.
        """
        self.table_options['banner_border'] = border

    def set_top_offset(self, offset_row):
        """Set where the data starts in the Excel sheet.
        """
        self.formats['offsets']['top'] = offset_row

    def freeze_panes(self, row=10, column=1):
        """Set where to freeze panes in the Excel file.
        """
        freeze = {
            "row": row,
            "col": column
        }
        self.formats['freeze_panes'] = freeze


    def set_row_colors(self, colors):
        """Set alternating row colors of the Excel spreadsheet
        Parameters
        ======
        colors: list
            list of colours that will be alternated across (supply two colours to get stripes)
        """
        if type(colors) == list:
            self.formats['row_colors'] = colors
        else:
            raise ValueError("The colours must be a list of one or more colours")

    def set_format(self, name, new_format):
        """Set format for a question type in the table

        Parameters
        ==========
        name: string (counts, percentage, stats, base)
        new_format: format dictionary, e.g. {'bold':True}
        """
        if name in self.formats:
            self.formats[name] = {**self.formats[name], **new_format}
        else:
            self.formats[name] = new_format

    @verify_no_tables
    def set_answer_format(self, answer_type, format):
        """ Set the format for an answer, e.g. Strongly agree, agree etc.

        """
        new_format = {0: {"format": format}}
        old_format = self.table_options['format'][answer_type]
        self.table_options['format'][answer_type] = {**old_format, **new_format}

    def set_question_format(self, answer_type, format):
        """Set the format rules for the question (line at the top of every table showing variable label)
        
        Parameters
        ==========
        answer_type : string (percentage, counts, stats, base)
        format : format dict, e.g. {'bold':True}
        """
        new_format = {"question": {"format": format}}
        old_format = self.table_options['format'][answer_type]
        self.table_options['format'][answer_type] = {**old_format, **new_format}

    @verify_no_tables
    def set_column_format_for_type(self, answer_type, column_index, format):
        """Set column format for answer type.

        Parameters
        ==========
         answer_type : string, percentage, counts, stats
         column_index : int
         format : format dict

        Example
        =======
        set_column_format_for_type('counts', 1, {'bold':True})
        """
        new_format = {column_index: {"format":format}}
        old_format = self.table_options['format'][answer_type]
        self.table_options['format'][answer_type] = {**old_format, **new_format}

    def _set_page_setup(self, key, value):
        if 'set_page_setup' not in self.formats:
            self.formats['set_page_setup'] = {}
        self.formats['set_page_setup'][key] = value

    def merge(self,  options):
        """ Merge two Option objects together, this one overriding the incoming one
        """
        self.table_options = self.merge_dict(self.table_options, copy.deepcopy(options.table_options))
        self.formats = self.merge_dict(self.formats, copy.deepcopy(options.formats))

    def merge_dict(self, source, destination):
        """ Merge two dictionaries by having the destination override the source, recursively.
        """
        for key, value in source.items():
            if isinstance(value, dict):
                # get node or create one
                node = destination.setdefault(key, {})
                self.merge_dict(value, node)
            else:
                destination[key] = value

        return destination


import json
import pandas as pd
import io
from functools import partial
import urllib.request

from difflib import SequenceMatcher
from .decorators import add_data, format_response, valid_params
from .tally import Tally

# these are the keys returned in results that indicate what should happen with the local dataset
VARIABLE_KEYS = ['meta', 'data']
META_VARIABLE_KEYS = ['meta']
NEW_DATASET_KEYS = ['dataset_meta', 'dataset_data']

class DataSet:
    """
    A class that wraps a dataset and has all the information needed to send to the API in order to perform the various tasks.

    Parameters
    ----------
      name: string
        Name for the dataset
    """
    dataset_type = None
    sav_data = None
    qp_meta = None
    qp_data = None
    tally = None

    def __init__(self, api_key=None, host='tally.datasmoothie.com', ssl=True):
        self.add_credentials(api_key=api_key, host=host, ssl=ssl)

    #@add_data
    def __getattr__(self, name):
        method = partial(self._call_tally, name)
        return method

    @add_data
    @format_response
    def _call_tally(self, api_endpoint, *args, **kwargs):
        data_params = kwargs.pop('data_params')
        files, payload = self.prepare_post_params(data_params, kwargs)
        response = self.tally.post_request('tally', api_endpoint, payload, files)
        if response.status_code == 404:
            return self._method_not_found_response(api_endpoint)
        json_dict = json.loads(response.content)
        json_dict = self._clean_error_response(json_dict)
        if self._has_keys(json_dict, VARIABLE_KEYS):
            self.add_column_to_data(json_dict['meta']['name'], json_dict['data'], json_dict['meta'])
            return
        if self._has_keys(json_dict, META_VARIABLE_KEYS):
            self.add_column_to_data(json_dict['meta']['name'], None, json_dict['meta'])
            return
        if self._has_keys(json_dict, NEW_DATASET_KEYS):
            self.qp_data = json_dict['dataset_data']
            self.qp_meta = json.dumps(json_dict['dataset_meta'])
            return
        return json_dict

    def _method_not_found_response(self, method):
        try:
            didyoumean = ""
            with urllib.request.urlopen("https://tally.datasmoothie.com/openapi/?format=openapi-json") as url:
                data = json.load(url)
                endpoints = [i.replace('/tally/', '')[:-1] for i in data['paths'].keys()]
                matches = []
                for endpoint in endpoints:
                    matches.append((endpoint, SequenceMatcher(None, method, endpoint).ratio()))
                matches.sort(key=lambda tup: tup[1],reverse=True)
                if matches[0][1] > 0.8:
                    didyoumean = matches[0][0]
        except:
            pass
        if len(didyoumean)>0:
            infotext = "Unknown method '{}'. Did you mean '{}'? See https://tally.datasmoothie.com for available methods.".format(method, didyoumean)
        else:
            infotext = "Unknown method '{}'. See https://tally.datasmoothie.com for available methods.".format(method)
        raise ValueError(infotext)


    def _clean_error_response(self, error_response):
        if 'error' in error_response:
            if 'payload' in error_response['error'] and error_response['error']['payload'] == {}:
                del error_response['error']['payload'] 
            if 'detailed_message' in error_response['error'] and error_response['error']['detailed_message'] == error_response['error']['message']:
                del error_response['error']['detailed_message']
        return error_response

    def _has_keys(self, response, required_keys):
        return all(elem in response.keys() for elem in required_keys)

    def _generate_functions_from_api(self):
        with urllib.request.urlopen("https://tally.datasmoothie.com/openapi/?format=openapi-json") as url:
            data = json.load(url)
        endpoints = [i.replace('/tally/', '')[:-1] for i in data['paths'].keys()]
        result = ""
        for endpoint in endpoints:
            available = dir(self)
            if endpoint not in available:
                result = result + self._endpoint_api_to_docstring(data, endpoint)

        return result

    def _endpoint_api_to_docstring(self, data, endpoint):
        docstring = "    \"\"\""
        path_info = data['paths']['/tally/{}/'.format(endpoint)]
        parameter_docstring = ""
        docstring = docstring + path_info['post']['description']
        # check if there are parameters
        if endpoint in data['components']['schemas'] and type(data['components']['schemas'][endpoint]['properties']['params']) == dict:
            param_properties = data['components']['schemas'][endpoint]['properties']['params']['properties']
            docstring = docstring + "\tParameters\n\t=====\n"
            for key in param_properties:
                docstring = docstring + """\t{} : {}\n \t\t{}\n""".format(key, param_properties[key]["type"], param_properties[key]["description"])
        docstring = docstring + '\t\"\"\"'

        function_string = """
            def {}(self, **kwargs):
            {}
            return self._call_tally('{}', **kwargs)
        """.format(endpoint, docstring, endpoint)
        print(function_string)
        return function_string


    def add_credentials(self, api_key=None, host='tally.datasmoothie.com', ssl=True):
        """
        Add your API key and what server it is authorized to connect to. Useful for on-prem installations and development.          
        """
        tally = Tally(api_key=api_key, host=host, ssl=ssl)
        self.tally = tally

    def use_spss(self, file_path):
        """
        Load SPSS file into memory as the file to send with all requests.

        Parameters
        ----------
            file_path: string
                Path to the sav file we want to use as our data.
        """
        # this is okay because the path format will be the same as the OS running this

        with open(file_path, mode='rb') as file:
            fileContent = file.read()
        
        payload={}
        files=[ ('spss',('Example Data (A).sav',io.BytesIO(fileContent),'application/x-spss-sav')) ]
        response = self.tally.post_request('tally', 'convert_data_to_csv_json', payload, files)
        result = json.loads(response.content)
        self.qp_meta = json.dumps(result['json'])
        self.qp_data = result['csv']
        self.dataset_type = 'quantipy'

    @add_data
    def write_spss(self, file_path, data_params, **kwargs):
        """
        Writes the dataset to an SPSS (sav) file.

        Parameters
        ----------
            file_path: string
                Path to the sav file to write.
        """
        payload = {}
        files, payload = self.prepare_post_params(data_params, {})
        response = self.tally.post_request('tally', 'convert_data_to_sav', payload, files)
        file = open(file_path, "wb")
        file.write(response.content)
        file.close()
        return response

    def write_quantipy(self, file_path_json, file_path_csv):
        """
        Write the case and meta data as Quantipy compatible json and csv files. 

        Parameters
        ----------
            file_path_json: string
                Path to the json meta data file to create.
            file_path_csv: string
                Path to the csv data file to create.
        """
        json_file = open(file_path_json, "w")
        n = json_file.write(self.qp_meta)
        json_file.close()
        csv_file = open(file_path_csv, "w")
        n = csv_file.write(self.qp_data)
        csv_file.close()

    def use_quantipy(self, meta_json, data_csv):
        """
        Load Quantipy meta and data files to this dataset.

        Parameters
        ----------
            meta_json: string
                Path to the json file we want to use as our meta data.
            data_csv: string
                Path to the csv file we want to use as our data file.
        """
        with open(meta_json) as json_file:
            self.qp_meta = json.dumps(json.load(json_file))
        self.qp_data = pd.read_csv(data_csv).to_csv()
        self.dataset_type = 'quantipy'

    def use_csv(self, csv_file):
        """
        Load CSV file into the dataset as the file to send with all requests.

        Parameters
        ----------
            csv_file: string
                Path to the CSV file we want to use as our data.
        """        
        csv = pd.read_csv(csv_file).to_csv()
        payload = {'csv': csv}
        response = self.tally.post_request('tally', 'convert_data_to_csv_json', payload)
        result = json.loads(response.content)
        self.qp_meta = json.dumps(result['json'])
        self.qp_data = result['csv']
        self.dataset_type = 'quantipy'

    def use_nebu(self, nebu_url):
        """
        Load remote Nebu/Enghouse file into the dataset as the file to send with all requests.

        Parameters
        ----------
            nebu_url: string
                Path to the Nebu data file we want to use as our data.
        """       
        payload = {
            'datasource': {
                'type' : 'Nebu',
                'nebu_url' : nebu_url
            }
        }
        response = self.tally.post_request('tally', 'convert_data_to_csv_json', payload)
        result = json.loads(response.content)
        self.qp_meta = json.dumps(result['json'])
        self.qp_data = result['csv']
        self.dataset_type = 'quantipy'

    def use_confirmit(self, source_projectid, source_idp_url, source_client_id, source_client_secret, source_public_url):
        """
        Load remote Forsta/Confirmit data into the dataset as the data to send with all requests.

        Parameters
        ----------
        source_projectid: string
            Project id of the survey
        source_idp_url: string
            IPD Url of the survey
        source_client_id: string
            Your client id
        source_client_secret: string
            Client secret (don't commit this to a repository)
        source_public_url: string
                Public url to source
        """               
        payload = {
            'datasource': {
                'type': 'Confirmit',
                'source_idp_url' : source_idp_url,
                'source_projectid': source_projectid,
                'source_client_id' : source_client_id,
                'source_client_secret' : source_client_secret,
                'source_public_site_url' : source_public_url
            },
        }
        response = self.tally.post_request('tally', 'convert_data_to_csv_json', payload)
        result = json.loads(response.content)
        self.qp_meta = json.dumps(result['json'])
        self.qp_data = result['csv']
        self.dataset_type = 'quantipy'

    def use_parquet(self, pq_data_filename, pq_meta_filename=None):

        with open(pq_data_filename, mode='rb') as file:
            fileContent_data = file.read()
        
        payload={}

        files=[ 
            ('pq', (pq_data_filename,io.BytesIO(fileContent_data), 'application/x-parquet')) 
        ]

        if pq_meta_filename is not None:
            with open(pq_meta_filename, mode='rb') as file:
                fileContent_meta = file.read()
            files.append(
                ('pq_meta',(pq_meta_filename,io.BytesIO(fileContent_meta),'text/plain')),
            )
        response = self.tally.post_request('tally', 'convert_data_to_csv_json', payload, files)
        result = json.loads(response.content)
        self.qp_meta = json.dumps(result['json'])
        self.qp_data = result['csv']
        self.dataset_type = 'quantipy'


    def use_unicom(self, mdd_filename, ddf_filename):
        with open(mdd_filename, mode='rb') as file:
            fileContent_mdd = file.read()

        with open(ddf_filename, mode='rb') as file:
            fileContent_ddf = file.read()
        
        payload={}

        files=[ 
            ('mdd',(mdd_filename,io.BytesIO(fileContent_mdd),'text/xml')),
            ('ddf', (ddf_filename,io.BytesIO(fileContent_ddf), 'application/x-sqlite3')) 
        ]
        response = self.tally.post_request('tally', 'convert_data_to_csv_json', payload, files)
        result = json.loads(response.content)
        self.qp_meta = json.dumps(result['json'])
        self.qp_data = result['csv']
        self.dataset_type = 'quantipy'

    def get_dataframe(self):
        return pd.read_csv(io.StringIO(self.qp_data))

    def prepare_post_params(self, data_params, params={}):
        # initialise the payload with our chosen data
        if 'binary_data' in data_params:
            files = data_params['binary_data']
            payload = {}
        else:
            files = None
            if type(data_params) == str:
                raise ValueError("The Tally client does not support non keyword arguments. For example, use dataset.crosstab(x='age'), not dataset.crosstab('age').")
            payload = data_params
        payload['params'] = params
        return (files, payload)

    @valid_params(['crosstabs', 'x', 'y', 'w', 'f', 'ci', 'stats', 'sig_level', 'decimals', 'base', 'painted', 'rebase', 'factors', 'format', 'add_format_column'])
    @add_data
    @format_response
    def crosstab(self, data_params=None, **kwargs):
        # initialise the payload with our chosen data
        # construct a crosstab option which references the dataset we are uploading with the request
        if 'crosstabs' not in kwargs.keys():
            kwargs['dataset'] = 'one'
            params = {"crosstabs":[kwargs]}
        else:
            crosstabs = kwargs['crosstabs']
            for ct in crosstabs:
                ct['dataset'] = 'one'
            params = {"crosstabs":crosstabs}
        files, payload = self.prepare_post_params(data_params, params)
        # the datasource will be a quantipy one, so we provide meta and data
        datasources={"one":{"meta":payload.pop('meta'), "data":payload.pop('data')}}
        payload['params']['datasources'] = datasources
        response = self.tally.post_request('tally', 'joined_crosstab', payload, files)
        json_dict = json.loads(response.content)
        if 'result' in json_dict.keys():
            return json_dict
        else:
            if 'message' in json_dict.keys():
                raise ValueError(json_dict['message'])
            else:
                raise ValueError("Crosstab returned no result.")

    @valid_params(['scheme', 'unique_key', 'variable', 'name'])
    @add_data
    def weight(self, data_params=None, **kwargs):
        files, payload = self.prepare_post_params(data_params, kwargs)
        response = self.tally.post_request('tally', 'weight', payload, files)
        json_dict = json.loads(response.content)
        if 'error' in json_dict:
            json_dict = self._clean_error_response(json_dict)
            return json_dict
        self.merge_column_to_data(kwargs['variable'], json_dict['weight_data'], json_dict['weight_var_meta'], kwargs['unique_key'])
        return json_dict

    def merge_column_to_data(self, name, data, new_meta, merge_on):
        df = pd.read_csv(io.StringIO(self.qp_data))
        new_series = pd.Series(data)
        new_series.index = new_series.index.astype(df[merge_on].dtype)
        new_series.name = name
        df = pd.merge(df, new_series, left_on=merge_on, right_on=new_series.index)
        self.qp_data = df.to_csv()
        meta = json.loads(self.qp_meta)
        meta['columns'][name] = new_meta
        self.qp_meta = json.dumps(meta)


    def add_column_to_data(self, name, data, new_meta):
        if data is not None:
            df = pd.read_csv(io.StringIO(self.qp_data))
            new_series = pd.Series(data)
            new_series.index = new_series.index.astype(int)
            df[name] = new_series
            self.qp_data = df.to_csv()
        meta = json.loads(self.qp_meta)
        meta['columns'][name] = new_meta
        self.qp_meta = json.dumps(meta)

    @add_data
    def convert_spss_to_csv_json(self, data_params=None):
        files, payload = self.prepare_post_params(data_params)
        response = self.tally.post_request('tally', 'convert_spss_to_csv_json', payload, files)
        json_dict = json.loads(response.content)
        result = {}
        result['csv'] = json_dict['csv']
        result['json'] = json_dict['json']
        return result

    @valid_params(['filename', 'sig_level', 'x', 'y', 'w', 'f', 'decimals'])
    @add_data
    def build_excel(self, data_params=None, filename=None, **kwargs):
        files, payload = self.prepare_post_params(data_params, kwargs)
        response = self.tally.post_request('tally', 'build_excel', payload, files)
        file = open(filename, "wb")
        file.write(response.content)
        file.close()
        return response

    @valid_params(['filename', 'dataframes', 'client_formats'])
    @add_data
    def build_excel_from_dataframes(self, data_params=None, filename=None, **kwargs):
        files, payload = self.prepare_post_params(data_params, kwargs)
        payload = kwargs
        response = self.tally.post_request('tally', 'build_excel_from_dataframes', payload, files)
        file = open(filename, "wb")
        file.write(response.content)
        file.close()
        return response

    @valid_params(['x', 'y', 'w', 'f', 'decimals', 'filename', 'powerpoint_template'])
    @add_data
    def build_powerpoint(self, data_params=None, filename=None, powerpoint_template=None, **kwargs):
        files, payload = self.prepare_post_params(data_params, kwargs)
        if files is None:
            files = {}
        files['pptx_template'] = (
            'template.pptx', 
            open(powerpoint_template, 'rb').read(), 
            'application/vnd.openxmlformats-officedocument.presentationml.presentation'
        )
        response = self.tally.post_request('tally', 'build_powerpoint', payload, files)
        file = open(filename, "wb")
        file.write(response.content)
        file.close()
        return response

    def band(self, **kwargs):
        """Group numeric data with band definitions treated as group text labels.

        Parameters
        =====
        name : string
            The column variable name keyed in _meta['columns'] that will be banded into summarized categories.
        bands : array
            The categorical bands to be used. Bands can be single numeric values or ranges.
        new_name : (string, default None)
            The created variable will be named '<name>_banded', unless a desired name is provided explicitly here.
        label : (string, default None)
            The created variable's text label will be identical to the originating one's passed in name, unless a desired label is provided explicitly here.
        text_key : (string, default None)
            Text key for text-based label information. Uses the DataSet.text_key information if not provided.
        """
        return self._call_tally('band', **kwargs)
        
    def compare(self, **kwargs):
        """Compares types, codes, values, question labels of two datasets.

        Parameters
        =====
        dataset : object
            (quantipy.DataSet instance). Test if all variables in the provided dataset are also in self and compare their metadata definitions.
        variables : str, array of str
            Check only these variables
        strict : (bool, default False)
            If True lower/ upper cases and spaces are taken into account.
        text_key : str, array of str
            Text key for text-based label information. Uses the DataSet.text_key information if not provided.
	    """
        return self._call_tally('compare', **kwargs)
        
    def convert_data_to_csv_json(self, **kwargs):
        """Converts data, either sent or from an external source to Quantipy CSV and JSON.

        The data to convert can be from parquet, SPSS, or UNICOM Intelligence (fka Dimensions) or a pure CSV exported from Excel.

        """
        return self._call_tally('convert_data_to_csv_json', **kwargs)
        
    def convert_data_to_sav(self, **kwargs):
        """Converts data, either sent or from an external source to an SPSS sav file.

        The data to convert can be from Quantipy, or UNICOM Intelligence (fka Dimensions) or a pure CSV exported from Excel.

        The sav files created do not support Quantipy's delimited set.

            """
        return self._call_tally('convert_data_to_sav', **kwargs)
        
    def copy(self, **kwargs):
        """Copy meta and case data of the variable defintion given per name.

        Parameters
        =====
        name : string
            The column variable name.
        suffix : string (default "rec")
            The new variable name will be constructed by suffixing the original `name` with `_suffix`, e.g. `age_rec`
        copy_data : boolean (default true)
            The new variable assumes the data of the original variable.
        slicer : dict
            If the data is copied it is possible to filter the data with a complex logic.
        copy_only : int or list of int, default None
            If provided, the copied version of the variable will only contain (data and) meta for the specified codes.
        copy_not : int or list of int, default None
            If provided, the copied version of the variable will contain (data and) meta for the all codes, except of the indicated.
        new_name : string
            If provided, the returned object will contain this new name instead of `name_suffix`
        """
        return self._call_tally('copy', **kwargs)
        
    def derive(self, **kwargs):
        """Create meta and recode case data by specifying derived category logics.

    Derived variables have their answer codes derived from other variables. Derived variables can either be
    multi-choice (called delimited set) or single-choice.

    A derived variable can be created from one variable, for example when a likert scale question
    has NETs added to it, or it can be created from multiple variables. When a derived variable is
    created from multiple variables, the user has to define how these variables should be used to create the 
    new one with by defining whether the new variable is an intersection or a union (logical expressions `and` or `or`).

    The conditional map has a list/array of of either three or four elements of following structures:

    #### 3 elements, type of logic not specified:
    [code, label, logic dictionary], e.g.:

    ``[1, "People who live in urban and sub-urban settings", {'locality': [1,2]}``

    #### 4 elements, type of logic speficied:
    [1, label, type of logic, logic_dictionary), e.g.:

    ``[1, "Men in urban and suburban locations", 
    'intersection', 
    {'gender': [1] 'locality': [1,2]))]``

    ``[2, "Women in urban and suburban locations", 'intersection', {'gender': [2] 'locality': [1,2]))``

    The logic types are 'union' and 'intersection'. If no logic type is specified, 'union' is used.
    `union` is equivalent to the logical expression `or` and `intersection` is equivalent to `and`.
    


        Parameters
        =====
        name : string
            The column variable name.
        label : string
            Name of the variable to show meta data for
        qtype : string
            The structural type of the data the meta describes (``int``, ``float``, ``single`` or ``delimited set``)
        cond_maps : array
            List of logic dictionaries that define how each answer and code is derived.
        cond_map : array (deprecated)
            List of "tuples", see documentatio above.
        """
        return self._call_tally('derive', **kwargs)
        
    def extend_values(self, **kwargs):
        """Add an answer/value and code to the list of answer/values/codes already in the meta data for the variable.

        Attempting to add already existing value codes or providing already
        present value texts will both raise invalid_arguments error!

        """
        return self._call_tally('extend_values', **kwargs)
        
    def feature_select(self, **kwargs):
        """Shows the variables that score the highest with a given ML features select algorithm

            	"""
        return self._call_tally('feature_select', **kwargs)
        
    def filter(self, **kwargs):
        """Filter the DataSet using a logical expression.

        Parameters
        =====
        alias : string
            Name of the filter
        condition : object
            An object that defines the filter logic
        """
        return self._call_tally('filter', **kwargs)
        
    def find(self, **kwargs):
        """Find variables by searching their names for substrings.

        Parameters
        =====
        str_tags : string or list of strings
            The strings tags to look for in the variable names. If not provided, the modules’ default global list of substrings from VAR_SUFFIXES will be used.
        suffixed : boolean (default false)
            If true, only variable names that end with a given string dequence will qualitfy.
        """
        return self._call_tally('find', **kwargs)
        
    def get_variable_text(self, **kwargs):
        """Return the variables text label information.

	"""
        return self._call_tally('get_variable_text', **kwargs)
        
    def hmerge(self, **kwargs):
        """Merge Quantipy datasets together by appending rows.
        This function merges two Quantipy datasets together,
        updating variables that exist in the left dataset and
        appending others.
        New variables will be appended in the order indicated by the
        'data file' set if found, otherwise they will be appended in
        alphanumeric order. This merge happens vertically (row-wise).

        Parameters
        =====
        dataset : object
            (quantipy.DataSet instance). Test if all variables in the provided dataset are also in self and compare their metadata definitions.
        on : str, default=None
            The column to use to identify unique rows in both datasets.
        left_on : str, default=None
            The column to use to identify unique in the left dataset.
        right_on : str, default=None
            The column to use to identify unique in the right dataset.
        row_id_name : str, default=None
            The named column will be filled with the ids indicated for each dataset, as per left_id/right_id/row_ids. If meta for the named column doesn't already exist a new column definition will be added and assigned a reductive-appropriate type.
        left_id : str, int, float, default=None
            Where the row_id_name column is not already populated for the dataset_left, this value will be populated.
        right_id : str, int, float, default=None
            Where the row_id_name column is not already populated for the dataset_right, this value will be populated.
        row_ids : array of (str, int, float), default=None
            When datasets has been used, this list provides the row ids that will be populated in the row_id_name column for each of those datasets, respectively.
        overwrite_text : bool, default=False
            If True, text_keys in the left meta that also exist in right meta will be overwritten instead of ignored.
        from_set : str, default=None
            Use a set defined in the right meta to control which columns are merged from the right dataset.
        uniquify_key : str, default=None
            An int-like column name found in all the passed DataSet objects that will be protected from having duplicates. The original version of the column will be kept under its name prefixed with 'original'.
        reset_index : bool, default=True
            If True pandas.DataFrame.reindex() will be applied to the merged dataframe.
        """
        return self._call_tally('hmerge', **kwargs)
        
    def joined_crosstab(self, **kwargs):
        """Does crosstab tabulation using the provided parameters, allowing for multiple
        datasources to be sent along with the request to run multiple crosstabs in one result. 
        Returns a json ecoded dataframe as a result. 
	    """
        return self._call_tally('joined_crosstab', **kwargs)
        
    def meta(self, **kwargs):
        """Shows the meta-data for a variable

        Parameters
        =====
        variable : string
            Name of the variable to show meta data for
        variables : array
            Name of multiple variables to show meta data for
        """
        return self._call_tally('meta', **kwargs)
        
    def recode(self, **kwargs):
        """Create a new or copied series from data, recoded using a mapper.

        This function takes a mapper of {key: logic} entries and injects the
        key into the target column where its paired logic is True. The logic
        may be arbitrarily complex and may refer to any other variable or
        variables in data. Where a pre-existing column has been used to
        start the recode, the injected values can replace or be appended to
        any data found there to begin with. The recoded data will always comply 
        with the column type indicated for the target column according to the meta. 

        #### Mapping example:
            recode_mapper = {
                1: {"$union":[{"$intersection": [{'locality': [3]}, {"gender":[1]}]},{"$intersection": [{'locality': [4]}, {"gender":[1]}]}]               },
                2: {"$intersection":[{'locality': [2]}, {'gender':[2]}]},
                3: {"$union":[{'locality': [1]}, {'gender':[1]}]},
                4: {'locality': [4]},
                5: {'locality': [5]},
                6: {'locality': [6]}
            }
        Logical functions are strings preceded with the symbol $ and logic can be nested at an arbitrary depth.


        Parameters
        =====
        target : string
            The variable name of the target of the recode.
        mapper : dict
            A mapper of {key: logic} entries.
        default : string
            The column name to default to in cases where unattended lists are given in your logic, where an auto-transformation of {key: list} to {key: {default: list}} is provided. Note that lists in logical statements are themselves a form of shorthand and this will ultimately be interpreted as: {key: {default: has_any(list)}}. 
        append : boolean
            Should the new recoded data be appended to values already found in the series? If False, data from series (where found) will overwrite whatever was found for that item instead.
        intersect : dict
            If a logical statement is given here then it will be used as an implied intersection of all logical conditions given in the mapper. For example, we could limit our mapper to males.
        initialize : str (default: None)
            Name of variable to use to populate the variable before the recode
        fillna : int
            If provided, can be used to fill empty/nan values.
        """
        return self._call_tally('recode', **kwargs)
        
    def remove_values(self, **kwargs):
        """Erase value codes safely from both meta and case data components.


    	"""
        return self._call_tally('remove_values', **kwargs)
        
    def set_value_texts(self, **kwargs):
        """Rename or add value texts in the ‘values’ object.

        This method works for array masks and column meta data.

        """
        return self._call_tally('set_value_texts', **kwargs)
        
    def set_variable_text(self, **kwargs):
        """Change the variable text for a named variable.

	"""
        return self._call_tally('set_variable_text', **kwargs)
        
    def sum(self, **kwargs):
        """Adds all values in each column and returns the sum for each column.
        Parameters
        =====
        new_variable_name : string
            Name for new variable that will contain the summarized amounts.
        variables : array
            The variables to sum. Only float or int types.
        """
        return self._call_tally('sum', **kwargs)
        
    def to_array(self, **kwargs):
        """Create a new variable grid (array) variable from two or more `single` variables with the same labels.

    	"""
        return self._call_tally('to_array', **kwargs)
        
    def to_delimited_set(self, **kwargs):
        """Create a new variable delimited set (multi choice) variable from two or more `single` variables.

    	"""
        return self._call_tally('to_delimited_set', **kwargs)
        
    def values(self, **kwargs):
        """Get a list of value texts and codes for a categorical variable, as a dictionary. 
        The method will return the codes in the data and the labels that apply to those codes
        in the chosen language (or the default language if no language is chosen).


        Parameters
        =====
        name : string
            Name of variable to fetch values/codes for.
        text_key : string (default None)
            The language key that should be used when taking labels from the meta data.
        include_variable_texts : boolean (default false)
            Include labels for the variable name in the results.
        """
        return self._call_tally('values', **kwargs)
        
    def variables(self, **kwargs):
        """
        Shows list of variables.

	    """
        return self._call_tally('variables', **kwargs)
        
    def vmerge(self, **kwargs):
        """
        Merge Quantipy datasets together by appending rows.
        This function merges two Quantipy datasets together, updating variables
        that exist in the left dataset and appending others. New variables
        will be appended in the order indicated by the 'data file' set if
        found, otherwise they will be appended in alphanumeric order. This
        merge happens vertically (row-wise).

        Parameters
        =====
        dataset : object
            (quantipy.DataSet instance). Test if all variables in the provided dataset are also in self and compare their metadata definitions.
        on : str, default=None
            The column to use to identify unique rows in both datasets.
        left_on : str, default=None
            The column to use to identify unique in the left dataset.
        right_on : str, default=None
            The column to use to identify unique in the right dataset.
        row_id_name : str, default=None
            The named column will be filled with the ids indicated for each dataset, as per left_id/right_id/row_ids. If meta for the named column doesn't already exist a new column definition will be added and assigned a reductive-appropriate type.
        left_id : str, int, float, default=None
            Where the row_id_name column is not already populated for the dataset_left, this value will be populated.
        right_id : str, int, float, default=None
            Where the row_id_name column is not already populated for the dataset_right, this value will be populated.
        row_ids : array of (str, int, float), default=None
            When datasets has been used, this list provides the row ids that will be populated in the row_id_name column for each of those datasets, respectively.
        overwrite_text : bool, default=False
            If True, text_keys in the left meta that also exist in right meta will be overwritten instead of ignored.
        from_set : str, default=None
            Use a set defined in the right meta to control which columns are merged from the right dataset.
        uniquify_key : str, default=None
            An int-like column name found in all the passed DataSet objects that will be protected from having duplicates. The original version of the column will be kept under its name prefixed with 'original'.
        reset_index : bool, default=True
            If True pandas.DataFrame.reindex() will be applied to the merged dataframe.
        """
        return self._call_tally('vmerge', **kwargs)
        


from .options import Options
import json
import pandas as pd
import copy

class Sheet:
    """ Represents a sheet in an Excel document.



    Parameters
    ----------
      name: string
        Name for the sheet. Will appear in table of contents.
      banner: list
        Variables to display across the top of the sheet.
      default_dataset: tally.DataSet
        The dataset to use for the data. The default is the build dataset, but sheets can have data from other datasets.
      parent : tally.Build
        The build object that owns the sheet.
    """

    def __init__(self, banner='@', default_dataset=None, name=None, parent=None):
        self.banner = banner
        self.tables = []
        self.default_dataset = default_dataset
        self.name = name
        self.options = Options(parent=self)
        self.parent = parent
        self.options._set_page_setup("write_string",{"row":2, "col":0, "string":name})

    def get_name(self):
        if self.name is not None:
            return self.name
        else:
            if len(self.dataframes)>0:
                name = self.dataframes[0].index.get_level_values(0)[0]
                if name == 'Base' or name == 'Unweighted base':
                    #check the format column if this is either a base or space row
                    return self.dataframes[0][~self.dataframes[0]['FORMAT'].str.contains('base|spacer', regex=True, case=False)].index[0][0]                       
                else:
                    return name
            else:
                return ""

    def add_table(self, stub, options={}, dataset=None):
        """ Add a table to the sheet

        Parameters
        ----------
        stub: dict
            Dictionary with information about what the table should contain. Options are
            x - the stub variable
            y - the banner variable
        options: dict
        dataset: tally.DataSet
        """
        if self.default_dataset and dataset is None:
            dataset = self.default_dataset
        merged_local_options = self.options.merge_dict(options, copy.deepcopy(self.options.table_options))
        merged_table_options = self.options.merge_dict(copy.deepcopy(merged_local_options), copy.deepcopy(self.parent.options.table_options))
        crosstab = {**{**merged_table_options['stub'], **stub}, **{'y':self.banner, 'add_format_column':True}}
        df = dataset.crosstab(
            crosstabs=[crosstab]
        )
        self.tables.append({"dataframe":df, "options":merged_table_options})

    def paint_dataframes(self):
        for table in self.tables:
            options = {**self.options.table_options, **table['options']}
            df = table['dataframe']
            # the format column should always have a cell_options key
            df['FORMAT'] = df['FORMAT'].apply(lambda x: json.dumps({**json.loads(x), **{'cell_format':{}}}))
            df = self.apply_table_options(df, options)
            table['dataframe'] = df
            table['dataframe'] = self._append_row_to_dataframe(table['dataframe'])
            if 'annotations' in self.options.table_options and self.options.table_options['annotations']:
                table = self._add_annotations(table)
            table['dataframe'] = self._append_row_to_dataframe(table['dataframe'])


    def apply_table_options(self, df, options):
        if 'banner_border' in options:
            df = self.add_banner_border(df)
        if 'title' in options:
            index = pd.MultiIndex.from_tuples([(' ',options['title']['text'])], names=['', ''])
            data_dict = {i:'' for i in df.columns}
            title_row = pd.DataFrame(data=data_dict, index=index, columns=df.columns)
            if 'format' in options['title']:
#                title_row['FORMAT'] = '{"type":"base", "cell_format":{"question":{"format":' + json.dumps(options['title']['format']) + '}}}'
                default_format = {'bold': True, 'align':'left'}
                cell_format = {**default_format, **options['title']['format']}
                title_row['FORMAT'] = '{"type":"base", "cell_format":{"0":{"format":' + json.dumps(cell_format) + '}}}'

            else:
                title_row['FORMAT'] = '{"type":"counts", "cell_format":{}}'
            df = pd.concat([title_row, df])
        if 'base' in options:
            df = self.apply_base_options(df, options['base'])
        if 'base_label' in options:
            df = df.rename(index={'Base':options['base_label']}, level=1)
        if 'unweighted_base_label' in options:
            df = df.rename(index={'Unweighted base':options['unweighted_base_label']}, level=1)
        if 'format' in options:
            df = self.apply_table_format_options(df, options['format'])
        if 'row_format' in options:
            df = self.set_row_format(df, options['row_format'])
        if 'font_name' in options:
            for i, col in enumerate(df.columns):
                self._set_column_format(df, i, {"font_name":options['font_name']})
        if 'font_size' in options:
            for i, col in enumerate(df.columns):
                self._set_column_format(df, i, {"font_size":options['font_size']})

        return df

    def apply_table_format_options(self, df, options):
        if 'base' in options:
            new_format = options['base']
            df = self.set_format_for_type(df, 'base', new_format)
        if 'percentage' in options:
            new_format = options['percentage']
            df = self.set_format_for_type(df, 'percentage', new_format)
        return df

    def set_row_format(self, df, format):
        def apply_row_format(x, new_format):
            current_format = json.loads(x)
            for i in list(range(df.shape[1])):
                if str(i) in current_format['cell_format']:
                    current_format['cell_format'][str(i)]['format'] = {**current_format['cell_format'][str(i)]['format'], **new_format['format']}
                else:
                    current_format['cell_format'][str(i)] = {'format':new_format['format']}
            return json.dumps(current_format)
        for row in format['rows']:
            df.iat[row, -1] = apply_row_format(df.iloc[row, -1], format)
        return df        

    def _set_column_format(self, df, col_index, format):
        def apply_column_format(x, col_index, new_format):
            current_format = json.loads(x)
            for i in col_index:
                if str(i) in current_format['cell_format']:
                    current_format['cell_format'][str(i)]['format'] = {**current_format['cell_format'][str(i)]['format'], **new_format}
                else:
                    current_format['cell_format'][str(i)] = {'format':new_format}
            return json.dumps(current_format)
        # users can send int and lists
        if type(col_index) != list:
            col_index = [col_index]
        for row in list(range(df.shape[0])):
            if 'spacer' not in json.loads(df.iloc[row, -1]):
                df.iat[row, -1] = apply_column_format(df.iloc[row, -1], col_index, format)
        return df     

    def set_format_for_type(self, df, type, format):
        """ Method to add/alter format json per cell or per question type.

        Parameters:
            type (string): can be base, count, percentage or any type returned by the Tally API
            format (dict): what gets merged with cell_format
        """
        def apply_format(x, new_format):
            old_format = json.loads(x)
            resulting_format = old_format
            resulting_format['cell_format'] = {**old_format['cell_format'], **new_format}
            return json.dumps(resulting_format)

        new_format = format

        # set index temporarily to (question, 0), (question, 1) ... to insure unique indices
        question = df.index.levels[0].values[0]
        old_index = df.index
        df.index = pd.MultiIndex.from_tuples([(question, i) for i in range(df.shape[0])])
        location = df[df['FORMAT'].str.contains(f'\"type\": \"{type}\"')].index
        if len(location) == 0:
            location = df[df['FORMAT'].str.contains(f'\"original_type\": \"{type}\"')].index
        loc_row_index = list(location.get_level_values(1))
        for row in loc_row_index:
            df.iat[row, -1] = apply_format(df.iloc[row,-1], new_format)
        df.index = old_index
        df.index = df.index.set_names(['Question', 'Values'])
        return df

    def add_banner_border(self, df):
        # when we decide where to put the vertical border, we drop all the levels we won't use for the border
        # if we have nested banners
        df_for_index_calculation = df.copy()
        nested_depth = len([i for i in df.columns.droplevel(0).names if i =='Values'])
        if nested_depth > 1:
            drop_how_many = (nested_depth - 1)*2
            df_for_index_calculation.columns = df.columns.droplevel(level=list(range(drop_how_many)))
        all_questions = df_for_index_calculation.columns.to_frame().index.to_series().reset_index()['Question']
        # find location of first unique question (this is the far left column in each subframe)
        all_questions = list(all_questions.drop_duplicates().iloc[:-1].index)
        # the question column is col 0 and we don't count that
        if 0 in all_questions:
            all_questions.remove(0)
        all_questions = [i+1 for i in all_questions]
        self._set_column_format(df, all_questions, {'left':1})
        return df

    def apply_base_options(self, df, option):
        if option == 'hide':
            if 'Unweighted base' in df.index.get_level_values(1):
                df = df.drop('Unweighted base', level=1)
            if 'Base' in df.index.get_level_values(1):            
                df = df.drop('Base', level=1)
        elif option == 'outside':
            bases = df.xs('Base', level=1)
            tuples = [("Base", "Base")]*bases.shape[0]
            bases.index = pd.MultiIndex.from_tuples(tuples)
            if 'Unweighted base' in df.index.get_level_values(1):
                unweighted_bases = df.xs('Unweighted base', level=1)
                tuples = [("Base", "Unweighted base")]*bases.shape[0]
                unweighted_bases.index = pd.MultiIndex.from_tuples(tuples)
                bases = pd.concat([bases, unweighted_bases])
            if 'Unweighted base' in df.index.get_level_values(1):
                df = df.drop('Unweighted base', level=1)
            if 'Base' in df.index.get_level_values(1):            
                df = df.drop('Base', level=1)
            bases = self._append_row_to_dataframe(bases)
            df = pd.concat([bases,df])

        return df

    def _append_row_to_dataframe(self, df, row_data=None, row_format={}, before=False):
        last_question = df.index.get_level_values(0)[-1]
        mi = pd.MultiIndex.from_tuples([(last_question, ' ')])
        if row_data is None:
            row_df = pd.DataFrame(columns=df.columns, data=[['']*df.shape[1]], index=mi)
        else:
            row_df = pd.DataFrame(columns=df.columns, data=[row_data], index=mi)
        row_df.iloc[-1, -1] = '{{"type":"counts", "cell_format":{}}}'.format(row_format)
        if before:
            df = pd.concat([row_df, df])
        else:
            df = pd.concat([df, row_df])
        return df

    def build_dataframes(self, build_options=None):
        if len(self.tables) == 0:
            raise Exception("Sheet {} has no tables".format(self.name))
        self.options.merge(build_options)
        self.options.set_question_format('percentage', {'text_wrap':True})
        self.options.set_question_format('counts', {'text_wrap':True})
        self.paint_dataframes()
        dataframes = [i['dataframe'] for i in self.tables]
        result_df = pd.concat(dataframes)
        return result_df

    def _add_annotations(self, table):
        annotations = []
        df = table['dataframe']
        if 'sig_level' in table['options']['stub']: 

            # [[''], ['A', 'B'], ['C', 'D', 'E']]
            sig_test_list = list(df.columns.to_frame()['Test-IDs'].groupby('Question').apply(list))
            # ['A/B', 'C/D/E/F/G']
            first_join = [i for i in [("/").join(i) for i in sig_test_list] if len(i)>0]
            # 'A/B - C/D/E/F/G'
            final_test_str = " - ".join(first_join)
            sig_levels = table['options']['stub']['sig_level']
            sig_levels = sorted(set(sig_levels), key=sig_levels.index)
            level_str = ", ".join(["{0:.0%}".format(i) for i in sig_levels])
            sig_str = "Proportions/Means: Columns Tested ({} risk level) - {}.".format(level_str, final_test_str)
            annotations.append(sig_str)
            annotations.append("Minimum base: 30 (**), Small base: 100 (*).")

        if 'f' in table['options']['stub']:
            # e.g. {'gender':[1], 'agecat':[1,2]}
            filter = table['options']['stub']['f']
            filter_str = ""
            meta_data = {}
            for filter_key in filter:
                meta_data[filter_key] = self.default_dataset.values(name=[filter_key], include_variable_texts=True, format='dict')
                values = filter[filter_key]
                # e.g. London, Reykjavik
                value_string = ", ".join([meta_data[filter_key]['values'][filter_key][str(i)] for i in values])
                question_string = meta_data[filter_key]["variable_texts"][filter_key]
                filter_str = filter_str + "{}: {}. ".format(question_string, value_string)
            annotations.append(filter_str)

        if 'w' in table['options']['stub']:
            weight_str = "Weighted with variable {}.".format(table['options']['stub']['w'])
            annotations.append(weight_str)

        if len(annotations) > 0:
            for annotation in annotations:
                new_row = ['']*df.shape[1]
                new_row[0] = annotation
                df = self._append_row_to_dataframe(df, new_row, format)
        
        table['dataframe'] = df
        return table





    def table_count(self):
        return len(self.tables)